/****************************************************************************/
/****************************************************************************/
/*                                                                          */
/*                               MAPMAKER.C                                 */
/*                                                                          */
/*        Software to generate maps based on digital planetary images       */
/*                (to be compiled with any ANSI C compiler)                 */
/*                                                                          */
/*                       (c) Christian Whler 2003                          */
/*                                                                          */
/*                    http://www.christian-woehler.de                       */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define PI 3.141592654
#define RMAX 400
#define DR 0.1
#define EPS 1.0e-04


double frac(double x)
{
  return(x-((int) x));
}

int iceil(double x)
{
  return(((int) x)+1);
}

int ifloor(double x)
{
  return((int) x);
}

/* bilinear interpolation of RGB image */
void bilin_intp(unsigned char ***img,int width,int height,double xh,double yh,unsigned char *rgb)
{
  int k;
  double hf;

  if((xh>0) && (xh<width-1) && (yh>0) && (yh<height-1))
  {
    for(k=0;k<3;k++) /* compute R, G, and B */
    {
      hf=frac(yh)*frac(xh)*img[iceil(yh)][iceil(xh)][k]+
        frac(yh)*(1.0-frac(xh))*img[iceil(yh)][ifloor(xh)][k]+
        (1.0-frac(yh))*frac(xh)*img[ifloor(yh)][iceil(xh)][k]+
        (1.0-frac(yh))*(1.0-frac(xh))*img[ifloor(yh)][ifloor(xh)][k];
      rgb[k]=(unsigned char) (hf+0.5);
    }
  }
}

/* allocate RGB image */
unsigned char *** malloc_rgb_image(int width,int height)
{
  int i,j,k;
  unsigned char ***data=NULL;
  
  /* allocate pointers to image rows */
  if((data=(unsigned char ***)malloc(height*sizeof(unsigned char **)))==NULL) return(NULL);
  for(i=0;i<height;i++)
  {
    /* allocate pointers to pixels in each image row */
    if((data[i]=(unsigned char **)malloc(width*sizeof(unsigned char *)))==NULL) return(NULL);
    /* allocate individual RGB pixels */
    for(j=0;j<width;j++)
    {
      if((data[i][j]=(unsigned char *)malloc(3*sizeof(unsigned char)))==NULL) return(NULL);
      /* initialize RGB pixels to zero */
      for(k=0;k<3;k++) data[i][j][k]=0;
    }
  }

  /* return allocated pointer */
  return(data);
}

/* free RGB image */
void free_rgb_image(unsigned char ***data,int width,int height)
{
  int i,j;
  
  /* free RGB pixels */
  for(i=0;i<height;i++) for(j=0;j<width;j++) free(data[i][j]);
  
  /* free image rows */
  for(i=0;i<height;i++) free(data[i]);
  
  /* free image */
  free(data);
}


int main(int argc,char **argv)
{
  int i,j,k,np;
  int width,height,map_width,map_height;
  double cm,b0,xlo,ylo,xlw,ylw,ithr,delta,gpp,beta;
  double cp,sp,lm,dl,dl1,dl2,dl3,bm,x3,y3,z3,x4,y4,x5,y5,p1,p2;
  double posangle,xq,yq,intens,h,xh,yh,ra,rp,cl,hrgb,rh[4];
  unsigned char ***img;
  unsigned char ***map;
  unsigned char rgb[3];
  char st[255];
  FILE *f;

  
  if(argc!=13)
  {
    fprintf(stderr,"\nMapMaker 1.2   (c) Christian Woehler 2003\n");
    fprintf(stderr,"                   http://www.christian-woehler.de\n\n");
    fprintf(stderr,"Syntax: %s <IMGFILE> <MAPFILE> <CM> <B0> <XLO> <YLO> <XLW> <YLW> <ITHR> <DELTA> <GPP> <BETA>\n\n",argv[0]);
    fprintf(stderr,"Command line parameters:\n");
    fprintf(stderr,"   IMGFILE: Name of input image (PPM format)\n");
    fprintf(stderr,"   MAPFILE: Name of output map image (PPM format)\n");
    fprintf(stderr,"   CM     : Central meridian (degrees)\n");
    fprintf(stderr,"   B0     : Inclination of planet's equator against line of sight (degrees)\n");
    fprintf(stderr,"   XLO    : Image X coordinate of eastern (right) limb\n");
    fprintf(stderr,"   YLO    : Image Y coordinate of eastern (right) limb\n");
    fprintf(stderr,"   XLW    : Image X coordinate of western (left) limb\n");
    fprintf(stderr,"   YLW    : Image Y coordinate of western (left) limb\n");
    fprintf(stderr,"   ITHR   : Intensity threshold to separate planet from background\n");
    fprintf(stderr,"   DELTA  : Angular range on disk to be mapped\n");
    fprintf(stderr,"   GPP    : Scale (degrees per pixel) of the map\n");
    fprintf(stderr,"   BETA   : Limb darkening coefficient (approx. 0...2)\n");
    fprintf(stderr,"\n");
    exit(1);
  }

  cm=atof(argv[3]);               /* central meridian                                       */
  b0=atof(argv[4])*PI/180.0;      /* inclination of planet's equator against line of sight  */
  xlo=atof(argv[5]);              /* image x coordinate of eastern (right) limb             */
  ylo=atof(argv[6]);              /* image y coordinate of eastern (right) limb             */
  xlw=atof(argv[7]);              /* image x coordinate of western (left) limb              */
  ylw=atof(argv[8]);              /* image y coordinate of western (left) limb              */
  ithr=atof(argv[9]);             /* intensity threshold to separate planet from background */
  delta=atof(argv[10])*PI/180.0;  /* angular range on disk to be mapped (deg)               */
  gpp=atof(argv[11]);             /* degrees per pixel in the map                           */
  beta=atof(argv[12]);            /* limb darkening coefficient                             */
  
  
  /******************************************/
  /* check if parameters are in valid range */
  /******************************************/
  if((cm<0.0) || (cm>360.0))
  {
    fprintf(stderr,"Central meridian <CM> is invalid (must be in range 0...360).\n");
    exit(1);
  }
  if((b0>0.5*PI) || (b0<-0.5*PI))
  {
    fprintf(stderr,"Inclination <B0> is invalid (must be in range -90...+90).\n");
    exit(1);
  }
  if((ithr<0.0) || (ithr>255.0))
  {
    fprintf(stderr,"Intensity threshold <ITHR> is invalid (must be in range 0...255).\n");
    exit(1);
  }
  if((delta<0.0) || (delta>0.5*PI))
  {
    fprintf(stderr,"Maximum latitude and longitude range <DELTA> is invalid (must be in range 0...90).\n");
    exit(1);
  }
  if(gpp<EPS)
  {
    fprintf(stderr,"Degrees per pixel value <GPP> is invalid (must be positive).\n");
    exit(1);
  }
  if(beta<0.0)
  {
    fprintf(stderr,"Limb darkening coefficient <BETA> is invalid (must be positive).\n");
    exit(1);
  }
  
  
  /**************************************/
  /* load input image file (PPM format) */
  /**************************************/
  if((f=fopen(argv[1],"rb"))==NULL)
  {
    fprintf(stderr,"Cannot open input image file %s.\n",argv[1]);
    exit(1);
  }
  /* read header */
  fgets(st,200,f);
  /* ignore comment */
  do
  {
    fgets(st,200,f);
  }
  while(st[0]=='#');
  /* read image width and height */
  sscanf(st,"%d %d",&width,&height);
  printf("width = %d   height = %d\n",width,height);
  /* read and ignore line containing number of greyvalues (255) */
  fgets(st,200,f);
  /* allocate RGB image */
  if((img=malloc_rgb_image(width,height))==NULL)
  {
    fprintf(stderr,"Memory allocation failed.\n");
    exit(1);
  }
  /* read RGB values */
  for(i=0;i<height;i++)
    for(j=0;j<width;j++)
      for(k=0;k<3;k++)
        img[i][j][k]=(unsigned char) fgetc(f);
  fclose(f);


  /**********************/
  /* allocate map image */
  /**********************/
  map_width=(int) (360.0/gpp+0.5);
  map_height=(int) (180.0/gpp+0.5);
  if((map=malloc_rgb_image(map_width,map_height))==NULL)
  {
    fprintf(stderr,"Memory allocation failed.\n");
    exit(1);
  }


  /****************************************************/
  /* compute centroid of the planet disk in the image */
  /****************************************************/
  xq=yq=0.0;
  np=0;
  for(i=0;i<height;i++)
  {
    for(j=0;j<width;j++)
    {
      intens=1.0*(img[i][j][0]+img[i][j][1]+img[i][j][2])/3.0;
      if(intens>ithr)
      {
        xq+=1.0*j;
        yq+=1.0*i;
        np++;
      }
    }
  }
  if(np==0)  /* no sufficiently bright pixels */
  {
    fprintf(stderr,"Image has no sufficiently bright pixels.\n");
    free_rgb_image(img,width,height);
    free_rgb_image(map,map_width,map_height);
    exit(1);
  }
  xq/=(1.0*np);
  yq/=(1.0*np);
  printf("centroid of planet disk: x = %5.1f   y = %5.1f\n",xq,yq);


  /**************************/
  /* compute position angle */
  /**************************/
  if(xlo<xlw) /* swap limb points */
  {
    h=xlo;xlo=xlw;xlw=h;
    h=ylo;ylo=ylw;ylw=h;
  }
  if((xlo-xq<EPS) || (xq-xlw<EPS))
  {
    fprintf(stderr,"Invalid limb points.\n");
    free_rgb_image(img,width,height);
    free_rgb_image(map,map_width,map_height);
    exit(1);
  }
  p1=atan((yq-ylo)/(xlo-xq));
  p2=atan((ylw-yq)/(xq-xlw));
  posangle=0.5*(p1+p2);
  printf("Position angle: %4.1f degrees\n",posangle*180.0/PI);


  /*********************************************************************/
  /* determine equatorial and polar radius of planet disk in the image */
  /*********************************************************************/
  for(i=0;i<4;i++)
  {
    cp=cos(posangle+0.5*i*PI);
    sp=sin(posangle+0.5*i*PI);
    xh=xq;
    yh=yq;
    rh[i]=0.0;
    do
    {
      rh[i]+=DR;
      xh=xq+rh[i]*cp;
      yh=yq-rh[i]*sp;
      bilin_intp(img,width,height,xh,yh,rgb);
      intens=1.0*(rgb[0]+rgb[1]+rgb[2])/3.0;
    }
    while((xh>0.0) && (xh<=width-1) && (yh>0.0) && (yh<=height-1) && (intens>ithr) && (rh[i]<=RMAX));
    if(intens>ithr)
    {
	  fprintf(stderr,"Limb of the planet disk could not be detected.\n");
      free_rgb_image(img,width,height);
      free_rgb_image(map,map_width,map_height);
      exit(1);
    }
  }
  ra=0.5*(rh[0]+rh[2]);            /* equatorial disk radius  */
  rp=0.5*(rh[1]+rh[3]);            /* polar disk radius       */
  if((ra/rp>0.98) && (ra/rp<1.02)) /* planet disk is circular */
  {
    ra=rp=0.5*(ra+rp);
  }
  printf("Equatorial and polar radius of planet disk in pixels: ra = %4.1f   rp = %4.1f\n",ra,rp);
  

  /***************/
  /* compute map */
  /***************/
  for(i=0;i<map_height;i++)
  {
    for(j=0;j<map_width;j++)
    {
      /* longitude in degrees */
      lm=1.0*(map_width-j)*gpp;
      /* longitude difference w. r. t. central meridian (take */
      /* into account 360 degree periodicity of longitude)    */
      dl1=cm-lm;
      dl2=cm-(lm-360.0);
      dl3=cm-(lm+360.0);
      dl=dl1;
      if(fabs(dl2)<fabs(dl)) dl=dl2;
      if(fabs(dl3)<fabs(dl)) dl=dl3;
      dl*=PI/180.0;
      /* latitude */
      bm=-1.0*(i-map_height/2)*gpp*PI/180.0;
      x3=cos(bm)*sin(dl);
      y3=sin(bm)*cos(b0)+cos(bm)*cos(dl)*sin(b0);
      z3=-sin(bm)*sin(b0)+cos(bm)*cos(dl)*cos(b0);
      if((fabs(x3)<1.0-EPS) && (fabs(y3)<1.0-EPS) && (z3>=0.0))
      { /* map pixel in valid range */
        if((fabs(y3)<=sin(delta)-EPS) && (fabs(x3)/sqrt(1.0-y3*y3)<=sin(delta)-EPS))
        {
          cl=(1.0+beta)/(1.0+beta*sqrt(1.0-x3*x3-y3*y3));
          x4=ra*x3;
          y4=rp*y3;
          x5=xq+(x4*cos(posangle)-y4*sin(posangle));
          y5=yq-(x4*sin(posangle)+y4*cos(posangle));
          bilin_intp(img,width,height,x5,y5,rgb);
          for(k=0;k<3;k++)
          {
            hrgb=cl*rgb[k];
            if(hrgb>255.0) hrgb=255.0;
            map[i][j][k]=(unsigned char) hrgb;
          }
        }
      }
    }
  }
  

  /************/
  /* save map */
  /************/
  if((f=fopen(argv[2],"wb"))==NULL)
  {
    fprintf(stderr,"Cannot open file %s for writing.\n",argv[2]);
    free_rgb_image(img,width,height);
    free_rgb_image(map,map_width,map_height);
    exit(1);
  }
  fprintf(f,"P6\n# map image created by MapMaker\n%d %d\n255\n",map_width,map_height);
  for(i=0;i<map_height;i++)
    for(j=0;j<map_width;j++)
      for(k=0;k<3;k++)
        fputc((int) map[i][j][k],f);
  fclose(f);
  printf("Map written to PPM image file %s\n",argv[2]);
  
  
  /*************************************/
  /* free original image and map image */
  /*************************************/
  free_rgb_image(img,width,height);
  free_rgb_image(map,map_width,map_height);
      
  return(0);
}
