/****************************************************************************/
/****************************************************************************/
/*                                                                          */
/*                         MAPMAKER_PHASE_POLAR.C                           */
/*                                                                          */
/*       Software to generate maps based on digital planetary images;       */
/*         planet disk must be illuminated by more than 50 percent          */
/*                (to be compiled with any ANSI C compiler)                 */
/*                                                                          */
/*                       (c) Christian Whler 2003                          */
/*                                                                          */
/*                    http://www.christian-woehler.de                       */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define PI 3.141592654
#define RMAX 400
#define DR 0.1
#define EPS 1.0e-04
#define SQR(x) (x)*(x)


double frac(double x)
{
  return(x-((int) x));
}

int iceil(double x)
{
  return(((int) x)+1);
}

int ifloor(double x)
{
  return((int) x);
}

/* bilinear interpolation of RGB image */
void bilin_intp(unsigned char ***img,int width,int height,double xh,double yh,unsigned char *rgb)
{
  int k;
  double hf;

  if((xh>0) && (xh<width-1) && (yh>0) && (yh<height-1))
  {
    for(k=0;k<3;k++) /* compute R, G, and B */
    {
      hf=frac(yh)*frac(xh)*img[iceil(yh)][iceil(xh)][k]+
        frac(yh)*(1.0-frac(xh))*img[iceil(yh)][ifloor(xh)][k]+
        (1.0-frac(yh))*frac(xh)*img[ifloor(yh)][iceil(xh)][k]+
        (1.0-frac(yh))*(1.0-frac(xh))*img[ifloor(yh)][ifloor(xh)][k];
      rgb[k]=(unsigned char) (hf+0.5);
    }
  }
}

/* allocate RGB image */
unsigned char *** malloc_rgb_image(int width,int height)
{
  int i,j,k;
  unsigned char ***data=NULL;
  
  /* allocate pointers to image rows */
  if((data=(unsigned char ***)malloc(height*sizeof(unsigned char **)))==NULL) return(NULL);
  for(i=0;i<height;i++)
  {
    /* allocate pointers to pixels in each image row */
    if((data[i]=(unsigned char **)malloc(width*sizeof(unsigned char *)))==NULL) return(NULL);
    /* allocate individual RGB pixels */
    for(j=0;j<width;j++)
    {
      if((data[i][j]=(unsigned char *)malloc(3*sizeof(unsigned char)))==NULL) return(NULL);
      /* initialize RGB pixels to zero */
      for(k=0;k<3;k++) data[i][j][k]=0;
    }
  }

  /* return allocated pointer */
  return(data);
}

/* free RGB image */
void free_rgb_image(unsigned char ***data,int width,int height)
{
  int i,j;
  
  /* free RGB pixels */
  for(i=0;i<height;i++) for(j=0;j<width;j++) free(data[i][j]);
  
  /* free image rows */
  for(i=0;i<height;i++) free(data[i]);

  /* free image */
  free(data);
}


int main(int argc,char **argv)
{
  int i,j,k;
  int width,height,map_width,map_height,nflag;
  double cm,b0,xln,yln,xls,yls,ithr,min_beta,delta,gpp,beta;
  double cp,sp,lm,dl,dl1,dl2,dl3,bm,xh0,yh0,x3,y3,z3,x4,y4,x5,y5;
  double posangle,xq,yq,intens,ra,rp,cl,hrgb,rh[2],xh[2],yh[2];
  unsigned char ***img;
  unsigned char ***map;
  unsigned char rgb[3];
  char st[255];
  FILE *f;


  if(argc!=15)
  {
    fprintf(stderr,"\nMapMaker_Phase_Polar 1.2   (c) Christian Woehler 2003\n");
    fprintf(stderr,"                           http://www.christian-woehler.de\n\n");
    fprintf(stderr,"Syntax: %s <IMGFILE> <MAPFILE> <CM> <B0> <XLN> <YLN> <XLS> <YLS> <ITHR> <N/S> <MIN_BETA> <DELTA> <GPP> <BETA>\n\n",argv[0]);
    fprintf(stderr,"Command line parameters:\n");
    fprintf(stderr,"   IMGFILE : Name of input image (PPM format)\n");
    fprintf(stderr,"   MAPFILE : Name of output map image (PPM format)\n");
    fprintf(stderr,"   CM      : Central meridian (degrees)\n");
    fprintf(stderr,"   B0      : Inclination of planet's equator against line of sight (degrees)\n");
    fprintf(stderr,"   XLN     : Image X coordinate of northern limb of planet disk\n");
    fprintf(stderr,"   YLN     : Image Y coordinate of northern limb of planet disk\n");
    fprintf(stderr,"   XLS     : Image X coordinate of southern limb of planet disk\n");
    fprintf(stderr,"   YLS     : Image Y coordinate of southern limb of planet disk\n");
    fprintf(stderr,"   ITHR    : Intensity threshold to separate planet from background\n");
    fprintf(stderr,"   N/S     : Compute map of northern ('N' or 'n') or southern ('S' or 's') polar region\n");
    fprintf(stderr,"   MIN_BETA: limit of latitude range\n");
    fprintf(stderr,"   DELTA   : Angular range on disk to be mapped\n");
    fprintf(stderr,"   GPP     : Scale (degrees per pixel) of the map\n");
    fprintf(stderr,"   BETA    : Limb darkening coefficient (approx. 0...2)\n");
    fprintf(stderr,"\n");
    exit(1);
  }

  cm=atof(argv[3]);                 /* central meridian                                       */
  b0=atof(argv[4])*PI/180.0;        /* inclination of planet's equator against line of sight  */
  xln=atof(argv[5]);                /* image x coordinate of northern limb of planet disk     */
  yln=atof(argv[6]);                /* image y coordinate of northern limb of planet disk     */
  xls=atof(argv[7]);                /* image x coordinate of southern limb of planet disk     */
  yls=atof(argv[8]);                /* image y coordinate of southern limb of planet disk     */
  ithr=atof(argv[9]);               /* intensity threshold to separate planet from background */
  min_beta=atof(argv[11]);          /* limit of latitude range                                */
  delta=atof(argv[12])*PI/180.0;    /* angular range on disk to be mapped (deg)               */
  gpp=atof(argv[13]);               /* degrees per pixel in the map                           */
  beta=atof(argv[14]);              /* limb darkening coefficient                             */


  /******************************************/
  /* check if parameters are in valid range */
  /******************************************/
  if((cm<0.0) || (cm>360.0))
  {
    fprintf(stderr,"Central meridian <CM> is invalid (must be in range 0...360).\n");
    exit(1);
  }
  if((b0>0.5*PI) || (b0<-0.5*PI))
  {
    fprintf(stderr,"Inclination <B0> is invalid (must be in range -90...+90).\n");
    exit(1);
  }
  if((ithr<0.0) || (ithr>255.0))
  {
    fprintf(stderr,"Intensity threshold <ITHR> is invalid (must be in range 0...255).\n");
    exit(1);
  }
  if((argv[10][0]=='N') || (argv[10][0]=='n'))
  {
    nflag=1;
  }
  else
  {
    if((argv[10][0]=='S') || (argv[10][0]=='s'))
    {
      nflag=-1;
    }
    else
    {
      fprintf(stderr,"N/S flag is invalid (must be N, n, S or s).\n");
      exit(1);
    }
  }

  if((min_beta<-90.0) || (min_beta>90.0))
  {
    fprintf(stderr,"Parameter MIN_BETA is invalid (must be in range 0...90).\n");
    exit(1);
  }
  if((delta<0.0) || (delta>0.5*PI))
  {
    fprintf(stderr,"Maximum latitude and longitude range <DELTA> is invalid (must be in range 0...90).\n");
    exit(1);
  }
  if(gpp<EPS)
  {
    fprintf(stderr,"Degrees per pixel value <GPP> is invalid (must be positive).\n");
    exit(1);
  }
  if(beta<0.0)
  {
    fprintf(stderr,"Limb darkening coefficient <BETA> is invalid (must be positive).\n");
    exit(1);
  }


  /**************************************/
  /* load input image file (PPM format) */
  /**************************************/
  if((f=fopen(argv[1],"rb"))==NULL)
  {
    fprintf(stderr,"Cannot open input image file %s.\n",argv[1]);
    exit(1);
  }
  /* read header */
  fgets(st,200,f);
  /* ignore comment */
  do
  {
    fgets(st,200,f);
  }
  while(st[0]=='#');
  /* read image width and height */
  sscanf(st,"%d %d",&width,&height);
  printf("width = %d   height = %d\n",width,height);
  /* read and ignore line containing number of greyvalues (255) */
  fgets(st,200,f);
  /* allocate RGB image */
  if((img=malloc_rgb_image(width,height))==NULL)
  {
    fprintf(stderr,"Memory allocation failed.\n");
    exit(1);
  }
  /* read RGB values */
  for(i=0;i<height;i++)
    for(j=0;j<width;j++)
      for(k=0;k<3;k++)
        img[i][j][k]=(unsigned char) fgetc(f);
  fclose(f);


  /**********************/
  /* allocate map image */
  /**********************/
  map_width=2*((int) ((90.0-min_beta)/gpp+0.5));
  map_height=map_width;
  if((map=malloc_rgb_image(map_width,map_height))==NULL)
  {
    fprintf(stderr,"Memory allocation failed.\n");
    exit(1);
  }


  /****************************************************/
  /* compute centroid of the planet disk in the image */
  /****************************************************/
  xq=0.5*(xln+xls);
  yq=0.5*(yln+yls);
  printf("Centroid of planet disk: x = %5.1f   y = %5.1f\n",xq,yq);


  /**************************/
  /* compute position angle */
  /**************************/
  posangle=atan(1.0*(xls-xln)/(yls-yln));
  printf("Position angle: %4.1f degrees\n",posangle*180.0/PI);


  /*********************************************************************/
  /* determine polar radius of planet disk in the image (phase !!!)    */
  /*********************************************************************/
  for(i=0;i<2;i++)
  {
    cp=cos(posangle+0.5*(2*i+1)*PI);
    sp=sin(posangle+0.5*(2*i+1)*PI);
    xh[i]=xq;
    yh[i]=yq;
    rh[i]=0.0;
    do
    {
      rh[i]+=DR;
      xh[i]=xq+rh[i]*cp;
      yh[i]=yq-rh[i]*sp;
      bilin_intp(img,width,height,xh[i],yh[i],rgb);
      intens=1.0*(rgb[0]+rgb[1]+rgb[2])/3.0;
    }
    while((xh[i]>0.0) && (xh[i]<=width-1) && (yh[i]>0.0) && (yh[i]<=height-1) && (intens>ithr) && (rh[i]<=RMAX));
    if(intens>ithr)
    {
	   fprintf(stderr,"Limb of the planet disk could not be detected.\n");
      free_rgb_image(img,width,height);
      free_rgb_image(map,map_width,map_height);
      exit(1);
    }
  }
  ra=rp=0.5*(rh[0]+rh[1]);            /* disk radius      */
  xq=0.5*(xh[0]+xh[1]);               /* refined centroid */
  yq=0.5*(yh[0]+yh[1]);               /* of planet disk   */
  printf("Polar radius of planet disk in pixels: rp = %4.1f\n",rp);
  printf("Refined centroid of planet disk: x = %5.1f   y = %5.1f\n",xq,yq);

  /***************/
  /* compute map */
  /***************/
  for(i=0;i<map_height;i++)
  {
    for(j=0;j<map_width;j++)
    {
      /* longitude in degrees */
      xh0=1.0*(j-map_width/2);
      yh0=1.0*(i-map_height/2);
      if(fabs(xh0)<EPS) xh0=EPS;
      lm=atan2(yh0,xh0)*180.0/PI;
      if(nflag==-1) lm=360.0-lm;
      /*if(fabs(yh0)<EPS) printf("   yh0=0: lm = %lf\n",lm);*/
      if(lm<0.0) lm+=360.0;
      /* longitude difference w. r. t. central meridian (take */
      /* into account 360 degree periodicity of longitude)    */
      dl1=lm-cm;
      dl2=(lm-360.0)-cm;
      dl3=(lm+360.0)-cm;
      /*dl1=-dl1;dl2=-dl2;dl3=-dl3;*/
      dl=dl1;
      if(fabs(dl2)<fabs(dl)) dl=dl2;
      if(fabs(dl3)<fabs(dl)) dl=dl3;
      dl*=PI/180.0;
      dl=-dl;
      /* latitude */
      bm=(90.0-gpp*sqrt(SQR(1.0*(j-map_width/2))+SQR(1.0*(i-map_height/2))))*PI/180.0;
      if(nflag==-1) bm=-bm;
      /* basic transformation equations */
      x3=cos(bm)*sin(dl);
      y3=sin(bm)*cos(b0)+cos(bm)*cos(dl)*sin(b0);
      z3=-sin(bm)*sin(b0)+cos(bm)*cos(dl)*cos(b0);
      if((fabs(x3)<1.0-EPS) && (fabs(y3)<1.0-EPS) && (z3>=0.0))
      { /* map pixel in valid range */
        /*if((fabs(y3)<=sin(delta)-EPS) && (fabs(x3)/sqrt(1.0-y3*y3)<=sin(delta)-EPS))*/
        if(asin(sqrt(x3*x3+y3*y3))<delta)
        {
          cl=(1.0+beta)/(1.0+beta*sqrt(1.0-x3*x3-y3*y3));
          x4=ra*x3;
          y4=rp*y3;
          x5=xq+(x4*cos(posangle)-y4*sin(posangle));
          y5=yq-(x4*sin(posangle)+y4*cos(posangle));
          bilin_intp(img,width,height,x5,y5,rgb);
          for(k=0;k<3;k++)
          {
            hrgb=cl*rgb[k];
            if(hrgb>255.0) hrgb=255.0;
            map[i][j][k]=(unsigned char) hrgb;
          }
        }
      }
    }
  }
  
  /* mark pole by green cross */
  if((map_width>=3) && (map_height>=3))
  {
    map[map_height/2+0][map_width/2+0][1]=255;
    map[map_height/2+0][map_width/2+0][0]=map[map_height/2+0][map_width/2+0][2]=0;
    map[map_height/2+1][map_width/2+0][1]=255;
    map[map_height/2+1][map_width/2+0][0]=map[map_height/2+1][map_width/2+0][2]=0;
    map[map_height/2-1][map_width/2+0][1]=255;
    map[map_height/2-1][map_width/2+0][0]=map[map_height/2-1][map_width/2+0][2]=0;
    map[map_height/2+0][map_width/2+1][1]=255;
    map[map_height/2+0][map_width/2+1][0]=map[map_height/2+0][map_width/2+1][2]=0;
    map[map_height/2+0][map_width/2-1][1]=255;
    map[map_height/2+0][map_width/2-1][0]=map[map_height/2+0][map_width/2-1][2]=0;
  }

  /************/
  /* save map */
  /************/
  if((f=fopen(argv[2],"wb"))==NULL)
  {
    fprintf(stderr,"Cannot open file %s for writing.\n",argv[2]);
    free_rgb_image(img,width,height);
    free_rgb_image(map,map_width,map_height);
    exit(1);
  }
  fprintf(f,"P6\n# map image created by MapMaker\n%d %d\n255\n",map_width,map_height);
  for(i=0;i<map_height;i++)
    for(j=0;j<map_width;j++)
      for(k=0;k<3;k++)
        fputc((int) map[i][j][k],f);
  fclose(f);
  printf("Map written to PPM image file %s\n",argv[2]);
  
  
  /*************************************/
  /* free original image and map image */
  /*************************************/
  free_rgb_image(img,width,height);
  free_rgb_image(map,map_width,map_height);
      
  return(0);
}
